from typing import Callable, Tuple

import aesara.tensor as at
import numpy as np
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.var import TensorVariable

from aehmc.integrators import IntegratorStateType

ProposalStateType = Tuple[
    IntegratorStateType, TensorVariable, TensorVariable, TensorVariable
]


def proposal_generator(kinetic_energy: Callable, divergence_threshold: float):
    def update(initial_energy, state):
        """Generate a new proposal from a trajectory state.

        The trajectory state records information about the position in the state
        space and corresponding potential energy. A proposal also carries a
        weight that is equal to the difference between the current energy and
        the previous one. It thus carries information about the previous state
        as well as the current state.

        Parameters
        ----------
        initial_energy:
            The initial energy.
        state:
            The new state.

        Return
        ------
        A tuple that contains the new proposal and a boolean that indicates
        whether the current transition is divergent.

        """
        q, p, potential_energy, _ = state
        new_energy = potential_energy + kinetic_energy(p)

        delta_energy = initial_energy - new_energy
        delta_energy = at.where(at.isnan(delta_energy), -np.inf, delta_energy)
        is_transition_divergent = at.abs_(delta_energy) > divergence_threshold

        weight = delta_energy
        log_p_accept = at.where(
            at.gt(delta_energy, 0),
            at.as_tensor(0, dtype=delta_energy.dtype),
            delta_energy,
        )

        return (state, new_energy, weight, log_p_accept), is_transition_divergent

    return update


# -------------------------------------------------------------------
#                     PROGRESSIVE SAMPLING
# -------------------------------------------------------------------


def progressive_uniform_sampling(
    srng: RandomStream, proposal: ProposalStateType, new_proposal: ProposalStateType
) -> ProposalStateType:
    """Uniform proposal sampling.

    Choose between the current proposal and the proposal built from the last
    trajectory state.

    Parameters
    ----------
    srng
        RandomStream object
    proposal
        The current proposal, it does not necessarily correspond to the
        previous state on the trajectory
    new_proposal
        The proposal built from the last trajectory state.

    Return
    ------
    Either the current or the new proposal.

    """
    state, energy, weight, _ = proposal
    new_state, new_energy, new_weight, _ = new_proposal

    p_accept = at.expit(new_weight - weight)
    do_accept = srng.bernoulli(p_accept)
    updated_proposal = maybe_update_proposal(do_accept, proposal, new_proposal)

    return updated_proposal


def progressive_biased_sampling(
    srng: RandomStream, proposal: ProposalStateType, new_proposal: ProposalStateType
) -> ProposalStateType:
    """Baised proposal sampling.

    Choose between the current proposal and the proposal built from the last
    trajectory state. Unlike uniform sampling, biased sampling favors new
    proposals. It thus biases the transition away from the trajectory's initial
    state.

    Parameters
    ----------
    srng
        RandomStream object
    proposal
        The current proposal, it does not necessarily correspond to the
        previous state on the trajectory
    new_proposal
        The proposal built from the last trajectory state.

    Return
    ------
    Either the current or the new proposal.

    """
    state, energy, weight, _ = proposal
    new_state, new_energy, new_weight, _ = new_proposal

    p_accept = at.clip(at.exp(new_weight - weight), 0.0, 1.0)
    do_accept = srng.bernoulli(p_accept)
    updated_proposal = maybe_update_proposal(do_accept, proposal, new_proposal)

    return updated_proposal


def maybe_update_proposal(
    do_accept: bool, proposal: ProposalStateType, new_proposal: ProposalStateType
) -> ProposalStateType:
    """Return either proposal depending on the boolean `do_accept`"""
    state, energy, weight, log_sum_p_accept = proposal
    new_state, new_energy, new_weight, new_log_sum_p_accept = new_proposal

    updated_weight = at.logaddexp(weight, new_weight)
    updated_log_sum_p_accept = at.logaddexp(log_sum_p_accept, new_log_sum_p_accept)

    updated_q = at.where(do_accept, new_state[0], state[0])
    updated_p = at.where(do_accept, new_state[1], state[1])
    updated_potential_energy = at.where(do_accept, new_state[2], state[2])
    updated_potential_energy_grad = at.where(do_accept, new_state[3], state[3])
    updated_energy = at.where(do_accept, new_energy, energy)

    return (
        (updated_q, updated_p, updated_potential_energy, updated_potential_energy_grad),
        updated_energy,
        updated_weight,
        updated_log_sum_p_accept,
    )



from typing import Callable, Dict, Tuple

import aesara
import aesara.tensor as at
import numpy as np
from aesara.ifelse import ifelse
from aesara.scan.utils import until
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.var import TensorVariable

from aehmc.integrators import IntegratorStateType
from aehmc.proposals import (
    ProposalStateType,
    progressive_biased_sampling,
    progressive_uniform_sampling,
    proposal_generator,
)

__all__ = ["static_integration", "dynamic_integration", "multiplicative_expansion"]

TerminationStateType = Tuple[
    TensorVariable, TensorVariable, TensorVariable, TensorVariable
]


# -------------------------------------------------------------------
#                       STATIC INTEGRATION
#
# This section contains algorithms that integrate the trajectory for
# a set number of integrator steps.
# -------------------------------------------------------------------


def static_integration(
    integrator: Callable,
    num_integration_steps: int,
) -> Callable:
    """Build a function that generates fixed-length trajectories.

    Parameters
    ----------
    integrator
       Function that performs one integration step.
    num_integration_steps
        The number of times we need to run the integrator every time the
        returned function is called.

    Returns
    -------
    A function that integrates the hamiltonian dynamics a
    `num_integration_steps` times.

    """

    def integrate(
        q_init: TensorVariable,
        p_init: TensorVariable,
        energy_init: TensorVariable,
        energy_grad_init: TensorVariable,
        step_size: TensorVariable,
    ) -> Tuple[IntegratorStateType, Dict]:
        """Generate a trajectory by integrating several times in one direction.

        Parameters
        ----------
        q_init
            The initial position.
        p_init
            The initial value of the momentum.
        energy_init
            The initial value of the potential energy.
        energy_grad_init
            The initial value of the gradient of the potential energy wrt the position.
        step_size
            The size of each step taken by the integrator.

        Returns
        -------
        A tuple with the last position, value of the momentum, potential energy,
        gradient of the potential energy wrt the position in a tuple as well as
        a dictionary that contains the update rules for all the shared variables
        updated in `scan`.

        """

        def one_step(q, p, potential_energy, potential_energy_grad):
            new_state = integrator(
                q, p, potential_energy, potential_energy_grad, step_size
            )
            return new_state

        [q, p, energy, energy_grad], updates = aesara.scan(
            fn=one_step,
            outputs_info=[
                {"initial": q_init},
                {"initial": p_init},
                {"initial": energy_init},
                {"initial": energy_grad_init},
            ],
            n_steps=num_integration_steps,
        )

        return (q[-1], p[-1], energy[-1], energy_grad[-1]), updates

    return integrate


# -------------------------------------------------------------------
#                       DYNAMIC INTEGRATION
#
# This section contains algorithms that determine the number of
# integrator steps dynamically using a termination criterion that
# is updated at every step.
# -------------------------------------------------------------------


def dynamic_integration(
    srng: RandomStream,
    integrator: Callable,
    kinetic_energy: Callable,
    update_termination_state: Callable,
    is_criterion_met: Callable,
    divergence_threshold: TensorVariable,
):
    """Integrate a trajectory and update the proposal sequentially in one direction
    until the termination criterion is met.

    Parameters
    ----------
    srng
        A RandomStream object that tracks the changes in a shared random state.
    integrator
        The symplectic integrator used to integrate the hamiltonian dynamics.
    kinetic_energy
        Function to compute the current value of the kinetic energy.
    update_termination_state
        Updates the state of the termination mechanism.
    is_criterion_met
        Determines whether the termination criterion has been met.
    divergence_threshold
        Value of the difference of energy between two consecutive states above which we say a transition is divergent.

    Returns
    -------
    A function that integrates the trajectory in one direction and updates a
    proposal until the termination criterion is met.

    """
    generate_proposal = proposal_generator(kinetic_energy, divergence_threshold)
    sample_proposal = progressive_uniform_sampling

    def integrate(
        previous_last_state: IntegratorStateType,
        direction: TensorVariable,
        termination_state: TerminationStateType,
        max_num_steps: TensorVariable,
        step_size: TensorVariable,
        initial_energy: TensorVariable,
    ):
        """Integrate the trajectory starting from `initial_state` and update
        the proposal sequentially until the termination criterion is met.

        Parameters
        ----------
        previous_last_state
            The last state of the previously integrated trajectory.
        direction int in {-1, 1}
            The direction in which to expand the trajectory.
        termination_state
            The state that keeps track of the information needed for the termination criterion.
        max_num_steps
            The maximum number of integration steps. The expansion will stop
            when this number is reached if the termination criterion has not
            been met.
        step_size
            The step size of the symplectic integrator.
        initial_energy
            Initial energy H0 of the HMC step (not to confused with the initial energy of the subtree)

        Returns
        -------
        A tuple with on the one hand: a new proposal (sampled from the states
        traversed while building the trajectory), the last state, the sum of the
        momenta values along the trajectory (needed for termination criterion),
        the updated termination state, the number of integration steps
        performed, a boolean that indicates whether the trajectory has diverged,
        a boolean that indicates whether the termination criterion was met. And
        on the other hand a dictionary that contains the update rules for the
        shared variables updated in the internal `Scan` operator.

        """

        def add_one_state(
            step,
            q_proposal,  # current proposal
            p_proposal,
            potential_energy_proposal,
            potential_energy_grad_proposal,
            energy_proposal,
            weight,
            sum_p_accept,
            q_last,  # state
            p_last,
            potential_energy_last,
            potential_energy_grad_last,
            momentum_sum: TensorVariable,  # sum of momenta
            momentum_ckpts,  # termination state
            momentum_sum_ckpts,
            idx_min,
            idx_max,
            trajectory_length,
        ):
            state = (
                q_last,
                p_last,
                potential_energy_last,
                potential_energy_grad_last,
            )
            termination_state = (momentum_ckpts, momentum_sum_ckpts, idx_min, idx_max)
            proposal = (
                (
                    q_proposal,
                    p_proposal,
                    potential_energy_proposal,
                    potential_energy_grad_proposal,
                ),
                energy_proposal,
                weight,
                sum_p_accept,
            )

            new_state = integrator(*state, direction * step_size)
            new_proposal, is_diverging = generate_proposal(initial_energy, new_state)
            sampled_proposal = sample_proposal(srng, proposal, new_proposal)

            new_momentum_sum = momentum_sum + new_state[1]
            new_termination_state = update_termination_state(
                termination_state, new_momentum_sum, new_state[1], step
            )
            has_terminated = is_criterion_met(
                new_termination_state, new_momentum_sum, new_state[1]
            )

            do_stop_integrating = is_diverging | has_terminated

            return (
                *sampled_proposal[0],
                sampled_proposal[1],
                sampled_proposal[2],
                sampled_proposal[3],
                *new_state,
                new_momentum_sum,
                *new_termination_state,
                trajectory_length + 1,
                is_diverging,
                has_terminated,
            ), until(do_stop_integrating)

        # We take one step away to start building the subtrajectory
        state = integrator(*previous_last_state, direction * step_size)
        proposal, is_diverging = generate_proposal(initial_energy, state)
        momentum_sum = state[1]
        termination_state = update_termination_state(
            termination_state,
            momentum_sum,
            state[1],
            0,
        )

        steps = at.arange(1, 1 + max_num_steps)
        traj, updates = aesara.scan(
            add_one_state,
            outputs_info=(
                *proposal[0],
                proposal[1],
                proposal[2],
                proposal[3],
                *state,
                momentum_sum,
                *termination_state,
                at.as_tensor(1, dtype=np.int32),
                None,
                None,
            ),
            sequences=steps,
        )

        new_proposal = (
            (traj[0][-1], traj[1][-1], traj[2][-1], traj[3][-1]),
            traj[4][-1],
            traj[5][-1],
            traj[6][-1],
        )
        new_state = (traj[7][-1], traj[8][-1], traj[9][-1], traj[10][-1])
        subtree_momentum_sum = traj[11][-1]
        new_termination_state = (traj[12][-1], traj[13][-1], traj[14][-1], traj[15][-1])
        trajectory_length = traj[-3][-1]  # defined as the number of steps taken
        is_diverging = traj[-2][-1] | is_diverging
        has_terminated = traj[-1][-1]

        return (
            new_proposal,
            new_state,
            subtree_momentum_sum,
            new_termination_state,
            trajectory_length,
            is_diverging,
            has_terminated,
        ), updates

    return integrate


def multiplicative_expansion(
    srng: RandomStream,
    trajectory_integrator: Callable,
    uturn_check_fn: Callable,
    max_num_expansions: TensorVariable,
):
    """Sample a trajectory and update the proposal sequentially
    until the termination criterion is met.

    The trajectory is sampled with the following procedure:
    1. Pick a direction at random;
    2. Integrate `num_step` steps in this direction;
    3. If the integration has stopped prematurely, do not update the proposal;
    4. Else if the trajectory is performing a U-turn, return current proposal;
    5. Else update proposal, `num_steps = num_steps ** rate` and repeat from (1).

    Parameters
    ----------
    srng
        A RandomStream object that tracks the changes in a shared random state.
    trajectory_integrator
        A function that runs the symplectic integrators and returns a new proposal
        and the integrated trajectory.
    uturn_check_fn
        Function used to check the U-Turn criterion.
    max_num_expansions
        The maximum number of trajectory expansions until the proposal is
        returned.

    """
    proposal_sampler = progressive_biased_sampling

    def expand(
        proposal,
        left_state,
        right_state,
        momentum_sum,
        termination_state,
        initial_energy,
        step_size,
    ):
        """Expand the current trajectory multiplicatively.

        At each step we draw a direction at random, build a subtrajectory starting
        from the leftmost or rightmost point of the current trajectory that is
        twice as long as the current trajectory.

        Once that is done, possibly update the current proposal with that of
        the subtrajectory.

        Parameters
        ----------
        proposal
            Current new state proposal.
        left_state
            The current leftmost state of the trajectory.
        right_state
            The current rightmost state of the trajectory.
        momentum_sum
            The current value of the sum of momenta along the trajectory.
        initial_energy
            Potential energy before starting to build the trajectory.
        step_size
            The size of each step taken by the integrator.

        """

        def expand_once(
            step,
            q_proposal,  # proposal
            p_proposal,
            potential_energy_proposal,
            potential_energy_grad_proposal,
            energy_proposal,
            weight,
            sum_p_accept,
            q_left,  # trajectory
            p_left,
            potential_energy_left,
            potential_energy_grad_left,
            q_right,
            p_right,
            potential_energy_right,
            potential_energy_grad_right,
            momentum_sum,  # sum of momenta along trajectory
            momentum_ckpts,  # termination_state
            momentum_sum_ckpts,
            idx_min,
            idx_max,
        ):

            left_state = (
                q_left,
                p_left,
                potential_energy_left,
                potential_energy_grad_left,
            )
            right_state = (
                q_right,
                p_right,
                potential_energy_right,
                potential_energy_grad_right,
            )
            proposal = (
                (
                    q_proposal,
                    p_proposal,
                    potential_energy_proposal,
                    potential_energy_grad_proposal,
                ),
                energy_proposal,
                weight,
                sum_p_accept,
            )
            termination_state = (
                momentum_ckpts,
                momentum_sum_ckpts,
                idx_min,
                idx_max,
            )

            do_go_right = srng.bernoulli(0.5)
            direction = at.where(do_go_right, 1.0, -1.0)
            start_state = ifelse(do_go_right, right_state, left_state)

            (
                new_proposal,
                new_state,
                subtree_momentum_sum,
                new_termination_state,
                subtrajectory_length,
                is_diverging,
                has_subtree_terminated,
            ), inner_updates = trajectory_integrator(
                start_state,
                direction,
                termination_state,
                2**step,
                step_size,
                initial_energy,
            )

            # Update the trajectory.
            # The trajectory integrator always integrates forward in time; we
            # thus need to switch the states if the other direction was picked.
            new_left_state = ifelse(do_go_right, left_state, new_state)
            new_right_state = ifelse(do_go_right, new_state, right_state)
            new_momentum_sum = momentum_sum + subtree_momentum_sum

            # Compute the pseudo-acceptance probability for the NUTS algorithm.
            # It can be understood as the average acceptance probability MC would give to
            # the states explored during the final expansion.
            acceptance_probability = at.exp(new_proposal[3]) / subtrajectory_length

            # Update the proposal
            #
            # We do not accept proposals that come from diverging or turning subtrajectories.
            # However the definition of the acceptance probability is such that the
            # acceptance probability needs to be computed across the entire trajectory.
            updated_proposal = (
                proposal[0],
                proposal[1],
                proposal[2],
                at.logaddexp(new_proposal[3], proposal[3]),
            )

            sampled_proposal = where_proposal(
                is_diverging | has_subtree_terminated,
                updated_proposal,
                proposal_sampler(srng, proposal, new_proposal),
            )

            # Check if the trajectory is turning and determine whether we need
            # to stop expanding the trajectory.
            is_turning = uturn_check_fn(
                new_left_state[1], new_right_state[1], new_momentum_sum
            )
            do_stop_expanding = is_diverging | is_turning | has_subtree_terminated

            return (
                (
                    *sampled_proposal[0],
                    sampled_proposal[1],
                    sampled_proposal[2],
                    sampled_proposal[3],
                    *new_left_state,
                    *new_right_state,
                    new_momentum_sum,
                    *new_termination_state,
                    acceptance_probability,
                    step + 1,
                    is_diverging,
                    is_turning,
                ),
                inner_updates,
                until(do_stop_expanding),
            )

        expansion_steps = at.arange(0, max_num_expansions)
        results, updates = aesara.scan(
            expand_once,
            outputs_info=(
                *proposal[0],
                proposal[1],
                proposal[2],
                proposal[3],
                *left_state,
                *right_state,
                momentum_sum,
                *termination_state,
                None,
                None,
                None,
                None,
            ),
            sequences=expansion_steps,
        )

        return results, updates

    return expand


def where_proposal(
    do_pick_left: bool,
    left_proposal: ProposalStateType,
    right_proposal: ProposalStateType,
) -> ProposalStateType:
    """Represents a switch between two proposals depending on a condition."""
    left_state, left_weight, left_energy, left_log_sum_p_accept = left_proposal
    right_state, right_weight, right_energy, right_log_sum_p_accept = right_proposal

    state = ifelse(do_pick_left, left_state, right_state)
    energy = at.where(do_pick_left, left_energy, right_energy)
    weight = at.where(do_pick_left, left_weight, right_weight)
    log_sum_p_accept = ifelse(
        do_pick_left, left_log_sum_p_accept, right_log_sum_p_accept
    )

    return (state, energy, weight, log_sum_p_accept)



from typing import Callable

import aesara.tensor as at
import numpy as np
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.var import TensorVariable

from aehmc import hmc, integrators, metrics
from aehmc.termination import iterative_uturn
from aehmc.trajectory import dynamic_integration, multiplicative_expansion

new_state = hmc.new_state


def new_kernel(
    srng: RandomStream,
    logprob_fn: Callable[[TensorVariable], TensorVariable],
    max_num_expansions: int = 10,
    divergence_threshold: int = 1000,
) -> Callable:
    """Build an iterative NUTS kernel.

    Parameters
    ----------
    srng
        A RandomStream object that tracks the changes in a shared random state.
    logprob_fn
        A function that returns the value of the log-probability density
        function of a chain at a given position.
    max_num_expansions
        The maximum number of times we double the length of the trajectory.
        Known as the maximum tree depth in most implementations.
    divergence_threshold
        The difference in energy above which we say the transition is
        divergent.

    Returns
    -------
    A function which, given a chain state, returns a new chain state.

    References
    ----------
    .. [0]: Phan, Du, Neeraj Pradhan, and Martin Jankowiak. "Composable effects
            for flexible and accelerated probabilistic programming in NumPyro." arXiv
            preprint arXiv:1912.11554 (2019).
    .. [1]: Lao, Junpeng, et al. "tfp. mcmc: Modern markov chain monte carlo
            tools built for modern hardware." arXiv preprint arXiv:2002.01184 (2020).

    """

    def potential_fn(x):
        return -logprob_fn(x)

    def step(
        q: TensorVariable,
        potential_energy: TensorVariable,
        potential_energy_grad: TensorVariable,
        step_size: TensorVariable,
        inverse_mass_matrix: TensorVariable,
    ):
        """Use the NUTS algorithm to propose a new state.

        Parameters
        ----------
        q
            The initial position.
        potential_energy
            The initial value of the potential energy.
        potential_energy_grad
            The initial value of the gradient of the potential energy wrt the position.
        step_size
            The step size used in the symplectic integrator
        inverse_mass_matrix
            One or two-dimensional array used as the inverse mass matrix that
            defines the euclidean metric.

        Returns
        -------
        A tuple that contains on the one hand: the new position, value of the
        potential energy, gradient of the potential energy, the acceptance
        probability, the number of times the trajectory expanded, whether the
        integration diverged, whether the trajectory turned on itself. On the
        other hand a dictionary that contains the update rules for the shared
        variables updated in the scan operator.

        """
        momentum_generator, kinetic_energy_fn, uturn_check_fn = metrics.gaussian_metric(
            inverse_mass_matrix
        )
        symplectic_integrator = integrators.velocity_verlet(
            potential_fn, kinetic_energy_fn
        )
        (
            new_termination_state,
            update_termination_state,
            is_criterion_met,
        ) = iterative_uturn(uturn_check_fn)
        trajectory_integrator = dynamic_integration(
            srng,
            symplectic_integrator,
            kinetic_energy_fn,
            update_termination_state,
            is_criterion_met,
            divergence_threshold,
        )
        expand = multiplicative_expansion(
            srng,
            trajectory_integrator,
            uturn_check_fn,
            max_num_expansions,
        )

        p = momentum_generator(srng)
        initial_state = (q, p, potential_energy, potential_energy_grad)
        initial_termination_state = new_termination_state(q, max_num_expansions)
        initial_energy = potential_energy + kinetic_energy_fn(p)
        initial_proposal = (
            initial_state,
            initial_energy,
            at.as_tensor(0.0, dtype=np.float64),
            at.as_tensor(-np.inf, dtype=np.float64),
        )
        result, updates = expand(
            initial_proposal,
            initial_state,
            initial_state,
            p,
            initial_termination_state,
            initial_energy,
            step_size,
        )

        # New MCMC proposal
        q_new = result[0][-1]
        potential_energy_new = result[2][-1]
        potential_energy_grad_new = result[3][-1]

        # Diagnostics
        is_turning = result[-1][-1]
        is_diverging = result[-2][-1]
        num_doublings = result[-3][-1]
        acceptance_probability = result[-4][-1]

        return (
            q_new,
            potential_energy_new,
            potential_energy_grad_new,
            acceptance_probability,
            num_doublings,
            is_turning,
            is_diverging,
        ), updates

    return step



from typing import Callable, Tuple

import aesara.tensor as at
from aesara import config
from aesara.tensor.var import TensorVariable

from aehmc import algorithms

WelfordAlgorithmState = Tuple[TensorVariable, TensorVariable, TensorVariable]


def covariance_adaptation(
    is_mass_matrix_full: bool = False,
) -> Tuple[Callable, Callable, Callable]:
    """Adapts the values in the mass matrix by computing the covariance
    between parameters.

    Parameters
    ----------
    is_mass_matrix_full
        When False the algorithm adapts and returns a diagonal mass matrix
        (default), otherwise adapts and returns a dense mass matrix.

    Returns
    -------
    init
        A function that initializes the step of the mass matrix adaptation.
    update
        A function that updates the state of the mass matrix.
    final
        A function that computes the inverse mass matrix based on the current
        state.
    """

    wc_init, wc_update, wc_final = algorithms.welford_covariance(is_mass_matrix_full)

    def init(
        n_dims: int,
    ) -> Tuple[TensorVariable, WelfordAlgorithmState]:
        """Initialize the mass matrix adaptation.

        Parameters
        ----------
        ndims
            The number of dimensions of the mass matrix, which corresponds to
            the number of dimensions of the chain position.

        Returns
        -------
        The initial value of the mass matrix and the initial state of the
        Welford covariance algorithm.

        """
        if n_dims == 0:
            inverse_mass_matrix = at.constant(1.0, dtype=config.floatX)
        elif is_mass_matrix_full:
            inverse_mass_matrix = at.eye(n_dims, dtype=config.floatX)
        else:
            inverse_mass_matrix = at.ones((n_dims,), dtype=config.floatX)

        wc_state = wc_init(n_dims)

        return inverse_mass_matrix, wc_state

    def update(
        position: TensorVariable, wc_state: WelfordAlgorithmState
    ) -> WelfordAlgorithmState:
        """Update the algorithm's state.

        Parameters
        ----------
        position
            The current position of the chain.
        wc_state
            Current state of Welford's algorithm to compute covariance.

        """
        new_wc_state = wc_update(position, *wc_state)
        return new_wc_state

    def final(wc_state: WelfordAlgorithmState) -> TensorVariable:
        """Final iteration of the mass matrix adaptation.

        In this step we compute the mass matrix from the covariance matrix computed
        by the Welford algorithm, applying the shrinkage used in Stan [1]_.

        Parameters
        ----------
        wc_state
            Current state of Welford's algorithm to compute covariance.

        Returns
        -------
        The value of the inverse mass matrix computed from the covariance estimate.

        References
        ----------
        .. [1]: Carpenter, B., Gelman, A., Hoffman, M. D., Lee, D., Goodrich, B.,
                Betancourt, M., ... & Riddell, A. (2017). Stan: A probabilistic programming
                language. Journal of statistical software, 76(1), 1-32.

        """
        _, m2, sample_size = wc_state
        covariance = wc_final(m2, sample_size)

        scaled_covariance = (sample_size / (sample_size + 5)) * covariance
        shrinkage = 1e-3 * (5 / (sample_size + 5))
        if covariance.ndim > 0:
            if is_mass_matrix_full:
                new_inverse_mass_matrix = (
                    scaled_covariance + shrinkage * at.identity_like(covariance)
                )
            else:
                new_inverse_mass_matrix = scaled_covariance + shrinkage
        else:
            new_inverse_mass_matrix = scaled_covariance + shrinkage

        return new_inverse_mass_matrix

    return init, update, final



from typing import Callable, Dict, Tuple

import aesara
import aesara.tensor as at
import numpy as np
from aesara.ifelse import ifelse
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.var import TensorVariable

import aehmc.integrators as integrators
import aehmc.metrics as metrics
import aehmc.trajectory as trajectory


def new_state(
    q: TensorVariable, logprob_fn: Callable
) -> Tuple[TensorVariable, TensorVariable, TensorVariable]:
    """Create a new HMC state from a position.

    Parameters
    ----------
    q
        The chain's position.
    logprob_fn
        The function that computes the value of the log-probability density
        function at any position.

    Returns
    -------
    A new HMC state, i.e. a tuple with the position, current value of the
    potential energy and gradient of the potential energy.

    """
    potential_energy = -logprob_fn(q)
    potential_energy_grad = aesara.grad(potential_energy, wrt=q)
    return q, potential_energy, potential_energy_grad


def new_kernel(
    srng: RandomStream,
    logprob_fn: Callable,
    divergence_threshold: int = 1000,
) -> Callable:
    """Build a HMC kernel.

    Parameters
    ----------
    srng
        A RandomStream object that tracks the changes in a shared random state.
    logprob_fn
        A function that returns the value of the log-probability density
        function of a chain at a given position.
    divergence_threshold
        The difference in energy above which we say the transition is
        divergent.

    Returns
    -------
    A kernel that takes the current state of the chain and that returns a new
    state.

    References
    ----------
    .. [0]: Neal, Radford M. "MCMC using Hamiltonian dynamics." Handbook of markov
            chain monte carlo 2.11 (2011): 2.


    """

    def potential_fn(x):
        return -logprob_fn(x)

    def step(
        q: TensorVariable,
        potential_energy: TensorVariable,
        potential_energy_grad: TensorVariable,
        step_size: TensorVariable,
        inverse_mass_matrix: TensorVariable,
        num_integration_steps: int,
    ) -> Tuple[
        Tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable, bool],
        Dict,
    ]:
        """Perform a single step of the HMC algorithm.

        Parameters
        ----------
        q
            The initial position.
        potential_energy
            The initial value of the potential energy.
        potential_energy_grad
            The initial value of the gradient of the potential energy wrt the position.
        step_size
            The step size used in the symplectic integrator
        inverse_mass_matrix
            One or two-dimensional array used as the inverse mass matrix that
            defines the euclidean metric.
        num_integration_steps
            The number of times we run the integrator at each step.

        Returns
        -------
        A tuple that contains on the one hand: the new position, value of the
        potential energy, gradient of the potential energy and acceptance
        propbability. On the other hand a dictionaruy that contains the update
        rules for the shared variables updated in the scan operator.

        """

        momentum_generator, kinetic_energy_fn, _ = metrics.gaussian_metric(
            inverse_mass_matrix
        )
        symplectic_integrator = integrators.velocity_verlet(
            potential_fn, kinetic_energy_fn
        )
        proposal_generator = hmc_proposal(
            symplectic_integrator,
            kinetic_energy_fn,
            num_integration_steps,
            divergence_threshold,
        )

        p = momentum_generator(srng)
        (
            q_new,
            _,
            potential_energy_new,
            potential_energy_grad_new,
            acceptance_probability,
            is_divergent,
        ), updates = proposal_generator(
            srng, q, p, potential_energy, potential_energy_grad, step_size
        )

        return (
            q_new,
            potential_energy_new,
            potential_energy_grad_new,
            acceptance_probability,
            is_divergent,
        ), updates

    return step


def hmc_proposal(
    integrator: Callable,
    kinetic_energy: Callable[[TensorVariable], TensorVariable],
    num_integration_steps: TensorVariable,
    divergence_threshold: int,
) -> Callable:
    """Builds a function that returns a HMC proposal.

    Parameters
    --------
    integrator
        The symplectic integrator used to integrate the hamiltonian dynamics.
    kinetic_energy
        The function used to compute the kinetic energy.
    num_integration_steps
        The number of times we need to run the integrator every time the
        returned function is called.
    divergence_threshold
        The difference in energy above which we say the transition is
        divergent.

    Returns
    -------
    A function that generates a new state for the chain.

    """
    integrate = trajectory.static_integration(integrator, num_integration_steps)

    def propose(
        srng: RandomStream,
        q: TensorVariable,
        p: TensorVariable,
        potential_energy: TensorVariable,
        potential_energy_grad: TensorVariable,
        step_size: TensorVariable,
    ) -> Tuple[
        Tuple[
            TensorVariable,
            TensorVariable,
            TensorVariable,
            TensorVariable,
            TensorVariable,
            TensorVariable,
        ],
        Dict,
    ]:
        """Use the HMC algorithm to propose a new state.

        Parameters
        ----------
        srng
            A RandomStream object that tracks the changes in a shared random state.
        q
            The initial position.
        potential_energy
            The initial value of the potential energy.
        potential_energy_grad
            The initial value of the gradient of the potential energy wrt the position.
        step_size
            The step size used in the symplectic integrator

        Returns
        -------
        A tuple that contains on the one hand: the new position, value of the
        potential energy, gradient of the potential energy and acceptance
        probability. On the other hand a dictionary that contains the update
        rules for the shared variables updated in the scan operator.

        """

        (
            new_q,
            new_p,
            new_potential_energy,
            new_potential_energy_grad,
        ), updates = integrate(q, p, potential_energy, potential_energy_grad, step_size)

        # flip the momentum to keep detailed balance
        flipped_p = -1.0 * new_p

        # compute transition-related quantities
        energy = potential_energy + kinetic_energy(p)
        new_energy = new_potential_energy + kinetic_energy(flipped_p)
        delta_energy = energy - new_energy
        delta_energy = at.where(at.isnan(delta_energy), -np.inf, delta_energy)
        is_transition_divergent = at.abs(delta_energy) > divergence_threshold

        p_accept = at.clip(at.exp(delta_energy), 0, 1.0)
        do_accept = srng.bernoulli(p_accept)
        (
            final_q,
            final_p,
            final_potential_energy,
            final_potential_energy_grad,
        ) = ifelse(
            do_accept,
            (new_q, flipped_p, new_potential_energy, new_potential_energy_grad),
            (q, p, potential_energy, potential_energy_grad),
        )

        return (
            final_q,
            final_p,
            final_potential_energy,
            final_potential_energy_grad,
            p_accept,
            is_transition_divergent,
        ), updates

    return propose



from typing import Callable, Tuple

import aesara.tensor as at
import aesara.tensor.slinalg as slinalg
from aesara.tensor.random.utils import RandomStream
from aesara.tensor.shape import shape_tuple
from aesara.tensor.var import TensorVariable


def gaussian_metric(
    inverse_mass_matrix: TensorVariable,
) -> Tuple[Callable, Callable, Callable]:
    r"""Hamiltonian dynamic on euclidean manifold with normally-distributed momentum.

    The gaussian euclidean metric is a euclidean metric further characterized
    by setting the conditional probability density :math:`\pi(momentum|position)`
    to follow a standard gaussian distribution. A Newtonian hamiltonian
    dynamics is assumed.

    Arguments
    ---------
    inverse_mass_matrix
        One or two-dimensional array corresponding respectively to a diagonal
        or dense mass matrix.

    Returns
    -------
    momentum_generator
        A function that generates a value for the momentum at random.
    kinetic_energy
        A function that returns the kinetic energy given the momentum.
    is_turning
        A function that determines whether a trajectory is turning back on
        itself given the values of the momentum along the trajectory.

    References
    ----------
    .. [1]: Betancourt, Michael. "A general metric for Riemannian manifold
            Hamiltonian Monte Carlo." International Conference on Geometric Science of
            Information. Springer, Berlin, Heidelberg, 2013.

    """

    if inverse_mass_matrix.ndim == 0:
        shape: Tuple = ()
        mass_matrix_sqrt = at.sqrt(at.reciprocal(inverse_mass_matrix))
        dot, matmul = lambda x, y: x * y, lambda x, y: x * y
    elif inverse_mass_matrix.ndim == 1:
        shape = (shape_tuple(inverse_mass_matrix)[0],)
        mass_matrix_sqrt = at.sqrt(at.reciprocal(inverse_mass_matrix))
        dot, matmul = at.dot, lambda x, y: x * y
    elif inverse_mass_matrix.ndim == 2:
        shape = (shape_tuple(inverse_mass_matrix)[0],)
        tril_inv = slinalg.cholesky(inverse_mass_matrix)
        identity = at.eye(*shape)
        mass_matrix_sqrt = slinalg.solve_lower_triangular(tril_inv, identity)
        dot, matmul = at.dot, at.dot
    else:
        raise ValueError(
            f"Expected a mass matrix of dimension 1 (diagonal) or 2, got {inverse_mass_matrix.ndim}"
        )

    def momentum_generator(srng: RandomStream) -> TensorVariable:
        norm_samples = srng.normal(0, 1, size=shape, name="momentum")
        momentum = matmul(mass_matrix_sqrt, norm_samples)
        return momentum

    def kinetic_energy(momentum: TensorVariable) -> TensorVariable:
        velocity = matmul(inverse_mass_matrix, momentum)
        kinetic_energy = 0.5 * dot(velocity, momentum)
        return kinetic_energy

    def is_turning(
        momentum_left: TensorVariable,
        momentum_right: TensorVariable,
        momentum_sum: TensorVariable,
    ) -> bool:
        """Generalized U-turn criterion.

        Parameters
        ----------
        momentum_left
            Momentum of the leftmost point of the trajectory.
        momentum_right
            Momentum of the rightmost point of the trajectory.
        momentum_sum
            Sum of the momenta along the trajectory.

        .. [1]: Betancourt, Michael J. "Generalizing the no-U-turn sampler to Riemannian manifolds." arXiv preprint arXiv:1304.1920 (2013).
        .. [2]: "NUTS misses U-turn, runs in cicles until max depth", Stan Discourse Forum
                https://discourse.mc-stan.org/t/nuts-misses-u-turns-runs-in-circles-until-max-treedepth/9727/46
        """
        velocity_left = matmul(inverse_mass_matrix, momentum_left)
        velocity_right = matmul(inverse_mass_matrix, momentum_right)

        rho = momentum_sum - (momentum_right + momentum_left) / 2
        turning_at_left = at.dot(velocity_left, rho) <= 0
        turning_at_right = at.dot(velocity_right, rho) <= 0

        is_turning = turning_at_left | turning_at_right

        return is_turning

    return momentum_generator, kinetic_energy, is_turning



from typing import Callable, Tuple

import aesara
from aesara.tensor.var import TensorVariable

IntegratorStateType = Tuple[
    TensorVariable, TensorVariable, TensorVariable, TensorVariable
]


def new_integrator_state(
    potential_fn: Callable, position: TensorVariable, momentum: TensorVariable
):
    """Create a new integrator state from the current values of the position and momentum."""
    potential_energy = potential_fn(position)
    potential_energy_grad = aesara.grad(potential_energy, position)
    return (position, momentum, potential_energy, potential_energy_grad)


def velocity_verlet(
    potential_fn: Callable[[TensorVariable], TensorVariable],
    kinetic_energy_fn: Callable[[TensorVariable], TensorVariable],
) -> Callable[
    [TensorVariable, TensorVariable, TensorVariable, TensorVariable, TensorVariable],
    IntegratorStateType,
]:
    """The velocity Verlet (or Verlet-Störmer) integrator.

    The velocity Verlet is a two-stage palindromic integrator [1]_ of the form
    (a1, b1, a2, b1, a1) with a1 = 0. It is numerically stable for values of
    the step size that range between 0 and 2 (when the mass matrix is the
    identity).

    Parameters
    ----------
    potential_fn
        A function that returns the potential energy of a chain at a given
        position.
    kinetic_energy_fn
        A function that returns the kinetic energy of a chain at a given
        position and a given momentum.

    References
    ----------
    .. [1]: Bou-Rabee, Nawaf, and Jesús Marıa Sanz-Serna. "Geometric
            integrators and the Hamiltonian Monte Carlo method." Acta Numerica 27
            (2018): 113-206.

    """
    a1 = 0
    b1 = 0.5
    a2 = 1 - 2 * a1

    def one_step(
        position: TensorVariable,
        momentum: TensorVariable,
        potential_energy: TensorVariable,
        potential_energy_grad: TensorVariable,
        step_size: TensorVariable,
    ) -> IntegratorStateType:

        momentum = momentum - b1 * step_size * potential_energy_grad

        kinetic_grad = aesara.grad(kinetic_energy_fn(momentum), momentum)
        position = position + a2 * step_size * kinetic_grad

        potential_energy = potential_fn(position)
        potential_energy_grad = aesara.grad(potential_energy, position)
        momentum = momentum - b1 * step_size * potential_energy_grad

        return (
            position,
            momentum,
            potential_energy,
            potential_energy_grad,
        )

    return one_step



from typing import Callable, Tuple

import aesara.tensor as at
from aesara.tensor.var import TensorVariable

from aehmc import algorithms


def dual_averaging_adaptation(
    target_acceptance_rate: TensorVariable = at.as_tensor(0.8),
    gamma: float = 0.05,
    t0: int = 10,
    kappa: float = 0.75,
) -> Tuple[Callable, Callable]:
    r"""Tune the step size to achieve a desired target acceptance rate.

    Let us note :math:`\epsilon` the current step size, :math:`\alpha_t` the
    metropolis acceptance rate at time :math:`t` and :math:`\delta` the desired
    aceptance rate. We define:

    .. math:
        H_t = \delta - \alpha_t

    the error at time t. We would like to find a procedure that adapts the
    value of :math:`\epsilon` such that :math:`h(x) =\mathbb{E}\left[H_t|\epsilon\right] = 0`
    Following [1]_, the authors of [2]_ proposed the following update scheme. If
    we note :math:``x = \log \epsilon` we follow:

    .. math:
        x_{t+1} \LongLeftArrow \mu - \frac{\sqrt{t}}{\gamma} \frac{1}{t+t_0} \sum_{i=1}^t H_i
        \overline{x}_{t+1} \LongLeftArrow x_{t+1}\\, t^{-\kappa}  + \left(1-t^\kappa\right)\overline{x}_t

    :math:`\overline{x}_{t}` is guaranteed to converge to a value such that
    :math:`h(\overline{x}_t)` converges to 0, i.e. the Metropolis acceptance
    rate converges to the desired rate.

    See reference [2]_ (section 3.2.1) for a detailed discussion.

    Parameters
    ----------
    initial_log_step_size:
        Initial value of the logarithm of the step size, used as an iterate in
        the dual averaging algorithm.
    target_acceptance_rate:
        Target acceptance rate.
    gamma
        Controls the speed of convergence of the scheme. The authors of [2]_ recommend
        a value of 0.05.
    t0: float >= 0
        Free parameter that stabilizes the initial iterations of the algorithm.
        Large values may slow down convergence. Introduced in [2]_ with a default
        value of 10.
    kappa: float in ]0.5, 1]
        Controls the weights of past steps in the current update. The scheme will
        quickly forget earlier step for a small value of `kappa`. Introduced
        in [2]_, with a recommended value of .75

    Returns
    -------
    init
        A function that initializes the state of the dual averaging scheme.
    update
        A function that updates the state of the dual averaging scheme.

    References
    ----------
    .. [1]: Nesterov, Yurii. "Primal-dual subgradient methods for convex
            problems." Mathematical programming 120.1 (2009): 221-259.
    .. [2]: Hoffman, Matthew D., and Andrew Gelman. "The No-U-Turn sampler:
            adaptively setting path lengths in Hamiltonian Monte Carlo." Journal
            of Machine Learning Research 15.1 (2014): 1593-1623.

    """
    da_init, da_update = algorithms.dual_averaging(gamma, t0, kappa)

    def update(
        acceptance_probability: TensorVariable,
        step: TensorVariable,
        log_step_size: TensorVariable,
        log_step_size_avg: TensorVariable,
        gradient_avg: TensorVariable,
        mu: TensorVariable,
    ) -> Tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable]:
        """Update the dual averaging adaptation state.

        Parameters
        ----------
        acceptance_probability
            The acceptance probability returned by the sampling algorithm.
        step
            The current step number.
        log_step_size
            The logarithm of the current value of the current step size.
        log_step_size_avg
            The average of the logarithm of the current step size.
        gradient_avg
            The current value of the averaged gradients.
        mu
            The points towards which iterates are shrunk.

        """
        gradient = target_acceptance_rate - acceptance_probability
        return da_update(
            gradient, step, log_step_size, log_step_size_avg, gradient_avg, mu
        )

    return da_init, update



import warnings
from typing import Dict, Iterable, List

import aesara.tensor as at
from aesara.graph.basic import Variable, ancestors
from aesara.graph.fg import FunctionGraph
from aesara.graph.opt_utils import optimize_graph
from aesara.tensor.basic_opt import ShapeFeature
from aesara.tensor.var import TensorVariable


def simplify_shapes(graphs: List[Variable]):
    """Simply the shape calculations in a list of graphs."""
    shape_fg = FunctionGraph(
        outputs=graphs,
        features=[ShapeFeature()],
        clone=False,
    )
    return optimize_graph(shape_fg).outputs


class RaveledParamsMap:
    """Maps a set of tensor variables to a vector of their raveled values."""

    def __init__(self, ref_params: Iterable[TensorVariable]):
        self.ref_params = tuple(ref_params)

        self.ref_shapes = [at.shape(p) for p in self.ref_params]
        self.ref_shapes = simplify_shapes(self.ref_shapes)

        self.ref_dtypes = [p.dtype for p in self.ref_params]

        ref_shapes_ancestors = set(ancestors(self.ref_shapes))
        uninferred_shape_params = [
            p for p in self.ref_params if (p in ref_shapes_ancestors)
        ]
        if any(uninferred_shape_params):
            # After running the shape optimizations, the graphs in
            # `ref_shapes` should not depend on `ref_params` directly.
            # If they do, it could imply that we need to sample parts of a
            # model in order to get the shapes/sizes of its parameters, and
            # that's a worst-case scenario.
            warnings.warn(
                "The following parameters need to be computed in order to determine "
                f"the shapes in this parameter map: {uninferred_shape_params}"
            )

        param_sizes = [at.prod(s) for s in self.ref_shapes]
        cumsum_sizes = at.cumsum(param_sizes)
        # `at.cumsum` doesn't return a tensor of a fixed/known size
        cumsum_sizes = [cumsum_sizes[i] for i in range(len(param_sizes))]
        self.slice_indices = list(zip([0] + cumsum_sizes[:-1], cumsum_sizes))
        self.vec_slices = [slice(*idx) for idx in self.slice_indices]

    def ravel_params(self, params: List[TensorVariable]) -> TensorVariable:
        """Concatenate the raveled vectors of each parameter."""
        return at.concatenate([at.atleast_1d(p).ravel() for p in params])

    def unravel_params(
        self, raveled_params: TensorVariable
    ) -> Dict[TensorVariable, TensorVariable]:
        """Unravel a concatenated set of raveled parameters."""
        return {
            k: v.reshape(s).astype(t)
            for k, v, s, t in zip(
                self.ref_params,
                [raveled_params[slc] for slc in self.vec_slices],
                self.ref_shapes,
                self.ref_dtypes,
            )
        }

    def __repr__(self):
        return f"{type(self).__name__}({self.ref_params})"



from typing import Callable, Tuple

import aesara
import aesara.tensor as at
import numpy as np
from aesara import config as config
from aesara.ifelse import ifelse
from aesara.scan.utils import until
from aesara.tensor.var import TensorVariable


def iterative_uturn(is_turning_fn: Callable) -> Tuple[Callable, Callable, Callable]:
    """U-Turn termination criterion to check reversiblity while expanding
    the trajectory.

    The code follows the implementation in Numpyro [0]_, which is equivalent to
    that in TFP [1]_.

    Parameter
    ---------
    is_turning_fn:
        A function which, given the new momentum and the sum of the momenta
        along the trajectory returns a boolean that indicates whether the
        trajectory is turning on itself. Depends on the metric.

    References
    ----------
    .. [0]: Phan, Du, Neeraj Pradhan, and Martin Jankowiak. "Composable effects
            for flexible and accelerated probabilistic programming in NumPyro." arXiv
            preprint arXiv:1912.11554 (2019).
    .. [1]: Lao, Junpeng, et al. "tfp. mcmc: Modern markov chain monte carlo
            tools built for modern hardware." arXiv preprint arXiv:2002.01184 (2020).

    """

    def new_state(
        position: TensorVariable, max_num_doublings: TensorVariable
    ) -> Tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable]:
        """Initialize the termination state

        Parameters
        ----------
        position
            Example chain position. Used to infer the shape of the arrays that
            store relevant momentam and momentum sums.
        max_num_doublings
            Maximum number of doublings allowed in the multiplicative
            expansion. Determines the maximum number of momenta and momentum
            sums to store.

        Returns
        -------
        A tuple that represents a new state for the termination criterion.

        """
        if position.ndim == 0:
            return (
                at.zeros(max_num_doublings, dtype=config.floatX),
                at.zeros(max_num_doublings, dtype=config.floatX),
                at.constant(0, dtype=np.int64),
                at.constant(0, dtype=np.int64),
            )
        else:
            num_dims = position.shape[0]
            return (
                at.zeros((max_num_doublings, num_dims), dtype=config.floatX),
                at.zeros((max_num_doublings, num_dims), dtype=config.floatX),
                at.constant(0, dtype=np.int64),
                at.constant(0, dtype=np.int64),
            )

    def update(
        state: Tuple,
        momentum_sum: TensorVariable,
        momentum: TensorVariable,
        step: TensorVariable,
    ) -> Tuple[TensorVariable, TensorVariable, TensorVariable, TensorVariable]:
        """Update the termination state.

        Parameters
        ----------
        state
            The current termination state
        momentum_sum
            The sum of all momenta along the trajectory
        momentum
            The current momentum on the trajectory
        step
            Current step in the trajectory integration (starting at 0)

        Return
        ------
        A tuple that represents the updated termination state.

        """
        momentum_ckpt, momentum_sum_ckpt, *_ = state
        idx_min, idx_max = ifelse(
            at.eq(step, 0), (state[2], state[3]), _find_storage_indices(step)
        )

        momentum_ckpt = at.where(
            at.eq(step % 2, 0),
            at.set_subtensor(momentum_ckpt[idx_max], momentum),
            momentum_ckpt,
        )
        momentum_sum_ckpt = at.where(
            at.eq(step % 2, 0),
            at.set_subtensor(momentum_sum_ckpt[idx_max], momentum_sum),
            momentum_sum_ckpt,
        )

        return (momentum_ckpt, momentum_sum_ckpt, idx_min, idx_max)

    def is_iterative_turning(
        state: Tuple, momentum_sum: TensorVariable, momentum: TensorVariable
    ) -> bool:
        """Check if any sub-trajectory is making a U-turn.

        If we visualize the trajectory as a balanced binary tree, the
        subtrajectories for which we need to check the U-turn criterion are the
        ones for which the current node is the rightmost node. The
        corresponding momenta and sums of momentum corresponding to the nodes
        for which we need to check the U-Turn criterion are stored between
        `idx_min` and `idx_max` in `momentum_ckpts` and `momentum_sum_ckpts`
        respectively.

        Parameters
        ----------
        state
            The current termination state
        momentum_sum
            The sum of all momenta along the trajectory
        momentum
            The current momentum on the trajectory
        step
            Current step in the trajectory integration (starting at 0)


        Return
        ------
        True if any sub-trajectory makes a U-turn, False otherwise.

        """
        momentum_ckpts, momentum_sum_ckpts, idx_min, idx_max = state

        def body_fn(i):
            subtree_momentum_sum = (
                momentum_sum - momentum_sum_ckpts[i] + momentum_ckpts[i]
            )
            is_turning = is_turning_fn(
                momentum_ckpts[i], momentum, subtree_momentum_sum
            )
            reached_max_iteration = at.lt(i - 1, idx_min)
            do_stop = at.any(is_turning | reached_max_iteration)
            return (i - 1, is_turning), until(do_stop)

        val, _ = aesara.scan(body_fn, outputs_info=(idx_max, None), n_steps=idx_max + 2)

        is_turning = val[1][-1]
        is_turning = at.where(
            at.lt(idx_max, idx_min), at.as_tensor(0, dtype="bool"), is_turning
        )

        return is_turning

    return new_state, update, is_iterative_turning


def _find_storage_indices(step: TensorVariable) -> Tuple[int, int]:
    """Find the indices between which the momenta and sums are stored.

    Parameter
    ---------
    step
        The current step in the trajectory integration.

    Return
    ------
    The min and max indices between which the values relevant to check the
    U-turn condition for the current step are stored.

    """

    def count_subtrees(nc0, nc1):
        do_stop = at.eq(nc0 & 1, 0)
        new_nc0 = nc0 // 2
        new_nc1 = nc1 + 1
        return (new_nc0, new_nc1), until(do_stop)

    (_, nc1), _ = aesara.scan(
        count_subtrees,
        outputs_info=(step, -1),
        n_steps=step + 1,
    )
    num_subtrees = nc1[-1]

    def find_idx_max(nc0, nc1):
        do_stop = at.eq(nc0, 0)
        new_nc0 = nc0 // 2
        new_nc1 = nc1 + (nc0 & 1)
        return (new_nc0, new_nc1), until(do_stop)

    init = at.as_tensor(step // 2, dtype=np.int64)
    init_nc1 = at.constant(0, dtype=np.int64)
    (nc0, nc1), _ = aesara.scan(
        find_idx_max, outputs_info=(init, init_nc1), n_steps=step + 1
    )
    idx_max = nc1[-1]

    idx_min = idx_max - num_subtrees + 1

    return idx_min, idx_max



from typing import List, Tuple

import aesara
import aesara.tensor as at
from aesara import config
from aesara.ifelse import ifelse
from aesara.tensor.shape import shape_tuple
from aesara.tensor.var import TensorVariable

from aehmc.mass_matrix import covariance_adaptation
from aehmc.step_size import dual_averaging_adaptation


def run(
    kernel,
    initial_state,
    num_steps=1000,
    *,
    is_mass_matrix_full=False,
    initial_step_size=at.as_tensor(1.0, dtype=config.floatX),
    target_acceptance_rate=0.80
):

    init_adapt, update_adapt, final_adapt = window_adaptation(
        num_steps, is_mass_matrix_full, initial_step_size, target_acceptance_rate
    )

    def one_step(
        warmup_step,
        q,  # chain state
        potential_energy,
        potential_energy_grad,
        step,  # dual averaging adaptation state
        log_step_size,
        log_step_size_avg,
        gradient_avg,
        mu,
        mean,  # mass matrix adaptation state
        m2,
        sample_size,
        step_size,  # parameters
        inverse_mass_matrix,
    ):
        chain_state = (q, potential_energy, potential_energy_grad)
        warmup_state = (
            (step, log_step_size, log_step_size_avg, gradient_avg, mu),
            (mean, m2, sample_size),
        )
        parameters = (step_size, inverse_mass_matrix)

        # Advance the chain by one step
        chain_state, inner_updates = kernel(*chain_state, *parameters)

        # Update the warmup state and parameters
        warmup_state, parameters = update_adapt(
            warmup_step, warmup_state, parameters, chain_state
        )

        return (
            chain_state[0],  # q
            chain_state[1],  # potential_energy
            chain_state[2],  # potential_energy_grad
            *warmup_state[0],
            *warmup_state[1],
            *parameters,
        ), inner_updates

    (da_state, mm_state), parameters = init_adapt(initial_state)

    warmup_steps = at.arange(0, num_steps)
    state, updates = aesara.scan(
        fn=one_step,
        outputs_info=(*initial_state, *da_state, *mm_state, *parameters),
        sequences=(warmup_steps,),
        name="window_adaptation",
    )

    last_chain_state = (state[0][-1], state[1][-1], state[2][-1])
    step_size = state[-2][-1]
    inverse_mass_matrix = state[-1][-1]

    return last_chain_state, (step_size, inverse_mass_matrix), updates


def window_adaptation(
    num_steps: int,
    is_mass_matrix_full: bool = False,
    initial_step_size: TensorVariable = at.as_tensor(1.0, dtype=config.floatX),
    target_acceptance_rate: TensorVariable = 0.80,
):
    mm_init, mm_update, mm_final = covariance_adaptation(is_mass_matrix_full)
    da_init, da_update = dual_averaging_adaptation(target_acceptance_rate)
    schedule = build_schedule(num_steps)

    schedule_stage = at.as_tensor([s[0] for s in schedule])
    schedule_middle_window = at.as_tensor([s[1] for s in schedule])

    def init(initial_chain_state: Tuple):
        if initial_chain_state[0].ndim == 0:
            num_dims = 0
        else:
            num_dims = shape_tuple(initial_chain_state[0])[0]
        inverse_mass_matrix, mm_state = mm_init(num_dims)

        da_state = da_init(initial_step_size)
        step_size = at.exp(da_state[1])

        warmup_state = (da_state, mm_state)
        parameters = (step_size, inverse_mass_matrix)
        return warmup_state, parameters

    def fast_update(p_accept, warmup_state, parameters):
        da_state, mm_state = warmup_state
        _, inverse_mass_matrix = parameters

        new_da_state = da_update(p_accept, *da_state)
        step_size = at.exp(new_da_state[1])

        return (new_da_state, mm_state), (step_size, inverse_mass_matrix)

    def slow_update(position, p_accept, warmup_state, parameters):
        da_state, mm_state = warmup_state
        _, inverse_mass_matrix = parameters

        new_da_state = da_update(p_accept, *da_state)
        new_mm_state = mm_update(position, mm_state)
        step_size = at.exp(new_da_state[1])

        return (new_da_state, new_mm_state), (step_size, inverse_mass_matrix)

    def slow_final(warmup_state):
        """We recompute the inverse mass matrix and re-initialize the dual averaging scheme at the end of each 'slow window'."""
        da_state, mm_state = warmup_state

        inverse_mass_matrix = mm_final(mm_state)

        if inverse_mass_matrix.ndim == 0:
            num_dims = 0
        else:
            num_dims = shape_tuple(inverse_mass_matrix)[0]
        _, new_mm_state = mm_init(num_dims)

        step_size = at.exp(da_state[1])
        new_da_state = da_init(step_size)

        warmup_state = (new_da_state, new_mm_state)
        parameters = (step_size, inverse_mass_matrix)
        return warmup_state, parameters

    def final(
        warmup_state: Tuple, parameters: Tuple
    ) -> Tuple[TensorVariable, TensorVariable]:
        da_state, _ = warmup_state
        _, inverse_mass_matrix = parameters
        step_size = at.exp(da_state[2])  # return stepsize_avg at the end
        return step_size, inverse_mass_matrix

    def update(step: int, warmup_state: Tuple, parameters: Tuple, chain_state: Tuple):
        position, _, _, p_accept, *_ = chain_state

        stage = schedule_stage[step]
        warmup_state, parameters = where_warmup_state(
            at.eq(stage, 0),
            fast_update(p_accept, warmup_state, parameters),
            slow_update(position, p_accept, warmup_state, parameters),
        )

        is_middle_window_end = schedule_middle_window[step]
        warmup_state, parameters = where_warmup_state(
            is_middle_window_end, slow_final(warmup_state), (warmup_state, parameters)
        )

        is_last_step = at.eq(step, num_steps - 1)
        parameters = ifelse(is_last_step, final(warmup_state, parameters), parameters)

        return warmup_state, parameters

    def where_warmup_state(do_pick_left, left_warmup_state, right_warmup_state):
        (left_da_state, left_mm_state), left_params = left_warmup_state
        (right_da_state, right_mm_state), right_params = right_warmup_state

        da_state = ifelse(do_pick_left, left_da_state, right_da_state)
        mm_state = ifelse(do_pick_left, left_mm_state, right_mm_state)
        params = ifelse(do_pick_left, left_params, right_params)

        return (da_state, mm_state), params

    return init, update, final


def build_schedule(
    num_steps: int,
    initial_buffer_size: int = 75,
    final_buffer_size: int = 50,
    first_window_size: int = 25,
) -> List[Tuple[int, bool]]:
    """Return the schedule for Stan's warmup.

    The schedule below is intended to be as close as possible to Stan's _[1].
    The warmup period is split into three stages:
    1. An initial fast interval to reach the typical set. Only the step size is
    adapted in this window.
    2. "Slow" parameters that require global information (typically covariance)
    are estimated in a series of expanding intervals with no memory; the step
    size is re-initialized at the end of each window. Each window is twice the
    size of the preceding window.
    3. A final fast interval during which the step size is adapted using the
    computed mass matrix.
    Schematically:

    ```
    +---------+---+------+------------+------------------------+------+
    |  fast   | s | slow |   slow     |        slow            | fast |
    +---------+---+------+------------+------------------------+------+
    ```

    The distinction slow/fast comes from the speed at which the algorithms
    converge to a stable value; in the common case, estimation of covariance
    requires more steps than dual averaging to give an accurate value. See _[1]
    for a more detailed explanation.

    Fast intervals are given the label 0 and slow intervals the label 1.

    Note
    ----
    It feels awkward to return a boolean that indicates whether the current
    step is the last step of a middle window, but not for other windows. This
    should probably be changed to "is_window_end" and we should manage the
    distinction upstream.

    Parameters
    ----------
    num_steps: int
        The number of warmup steps to perform.
    initial_buffer: int
        The width of the initial fast adaptation interval.
    first_window_size: int
        The width of the first slow adaptation interval.
    final_buffer_size: int
        The width of the final fast adaptation interval.

    Returns
    -------
    A list of tuples (window_label, is_middle_window_end).

    References
    ----------
    .. [1]: Stan Reference Manual v2.22 Section 15.2 "HMC Algorithm"

    """
    schedule = []

    # Give up on mass matrix adaptation when the number of warmup steps is too small.
    if num_steps < 20:
        schedule += [(0, False)] * (num_steps - 1)
    else:
        # When the number of warmup steps is smaller that the sum of the provided (or default)
        # window sizes we need to resize the different windows.
        if initial_buffer_size + first_window_size + final_buffer_size > num_steps:
            initial_buffer_size = int(0.15 * num_steps)
            final_buffer_size = int(0.1 * num_steps)
            first_window_size = num_steps - initial_buffer_size - final_buffer_size

        # First stage: adaptation of fast parameters
        schedule += [(0, False)] * (initial_buffer_size - 1)
        schedule.append((0, False))

        # Second stage: adaptation of slow parameters in successive windows
        # doubling in size.
        final_buffer_start = num_steps - final_buffer_size

        next_window_size = first_window_size
        next_window_start = initial_buffer_size
        while next_window_start < final_buffer_start:
            current_start, current_size = next_window_start, next_window_size
            if 3 * current_size <= final_buffer_start - current_start:
                next_window_size = 2 * current_size
            else:
                current_size = final_buffer_start - current_start
            next_window_start = current_start + current_size
            schedule += [(1, False)] * (next_window_start - 1 - current_start)
            schedule.append((1, True))

        # Last stage: adaptation of fast parameters
        schedule += [(0, False)] * (num_steps - 1 - final_buffer_start)
        schedule.append((0, False))

    return schedule
